#!/usr/bin/env python
# coding:utf-8

"""
sDSM (small Disk Space Monitor)

This file is part of sDSM.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = "Christian Schwamborn <cs@mail.architektur.tu-darmstadt.de>"
__copyright__ = "Copyright (c) 2016 Christian Schwamborn"
__license__ = "GPL v.3"


# TODO:	in verify data: use external function to verify option values
# TODO: in setData: check if the returned type of an option matches verify data
# TODO: write some write/export config function


import ConfigParser, re

msg0100 = 'config file \'{0}\' not found or inaccessible'
msg0101 = 'config file \'{0}\' successfully loaded'
msg0102 = 'adding section \'{0}\' to config'
msg0103 = 'section \'{0}\' option \'{1}\' not found in defaults'
msg0104 = 'setting default for option \'{0} = {1}\''
msg0105 = 'no verify data found for option \'{0}\''
msg0106 = 'internal error v_class should be one of \'str\' \'bool\' \'int\' \'float\' not: \'{0}\''
msg0107 = 'get missing configuration from defaults'
msg0108 = 'set verify data to: \'{0}\''
msg0109 = 'retrieved option \'{0}={1}\' in section \'{2}\' from config'
msg0110 = 'trying to transform \'{0}\' to a dict'
msg0111 = 'adding new element to dict: \'{0}\''
msg0112 = 'section \'{0}\' option \'{1}\' not found in config returning \'None\' as value'
msg0113 = 'unable to transform string to bool: \'{0}\''
msg0114 = 'unable to transform string to int or float: \'{0}\''
msg0115 = 'adding new element to list: \'{0}\''
msg0116 = 'trying to transform \'{0}\' to a list'
msg0117 = 'unable to transform \'{0}\' to a valid dict'
msg0118 = 'section \'{0}\' not found in defaults'
msg0119 = 'configuration error in section \'{0}\' option \'{1}\''
msg0120 = 'returning config sections: \'{0}\''
msg0121 = 'adding section \'{0}\' to defaults'
msg0122 = 'flushed verify data'
msg0123 = 'merged into verify data: \'{0}\''
msg0124 = 'retrieved options for section \'{0}\': \'{1}\''
msg0125 = 'set config data to: \'{0}\''
msg0126 = 'returning defaults data: {0}'
msg0127 = 'skipping existing option: \'{0}\''
msg0128 = 'get defaults for missing options in section: \'{0}\''
msg0129 = 'returning verify data: {0}'
msg0130 = 'returning config data: {0}'
msg0131 = 'option deleted from verify data: \'{0}\''
msg0132 = 'failed to delete option from verify data: \'{0}\''
msg0133 = 'failed to merge into verify data: \'{0}\''
msg0134 = 'set defaults data to: \'{0}\''
msg0135 = 'flushed defaults data'
msg0136 = 'merged into defaults data: \'{0} : {1}\''
msg0137 = 'failed to merge into defaults data: \'{0} : {1}\''
msg0138 = 'option deleted from defaults data: \'{0} : {1}\''
msg0139 = 'failed to delete option from defaults data: \'{0} : {1}\''
msg0140 = 'section deleted from defaults data: \'{0}\''
msg0141 = 'failed to delete section from defaults data: \'{0}\''
msg0142 = 'section \'{0}\' not found in config returning an empty options list'
msg0143 = 'section \'{0}\' not found in config returning \'None\' as value'
msg0144 = 'section \'{0}\' option \'{1}\' syntax error: {2}'
msg0145 = 'setting list delimiter to: \'{0}\''
msg0146 = 'trying to match \'{0}\' against a regex'
msg0147 = 'regex error while testing \'{0}\''
msg0148 = 'value \'{0}\' not in validation list: {1}'
msg0149 = 'value \'{0}\' found in validation list: {1}'
msg0150 = 'unable to match \'{0}\' against a regex'


class ConfigError(Exception):

	def __init__(self, *args):

		Exception.__init__(self, args) 


	def __str__(self):

		return str(self.args)



class Config():

	"""The verify dictionary
	is a dict with cfg options as key and lists as values.

	A single value option is a list with one (or two) item(s) of type sting:
	'str'|'int'|'float'|'bool'|'re'
	Except for 'bool' a second item can exist: In case of 're' it's an re
	match object, in the other cases it's list of values to validate against.

	A list value option requires a list of two (or three) items:
	'list', ('str'|'int'|'float'|'bool'|'re')
	meaning: A list of i.e. 'str'
	The same as for single value options applies here for the additional
	list element.

	A dict value option requires a list of three (or four) items:
	'dict', ('str'|'int'|'float'), ('str'|'int'|'float'|'bool'|'re')
	meaning: A dict with i.e. strings as keys and i.e. integers as values
	The same as for single value options applies here for the additional
	list element.
	"""

	def __init__(self, cf, logger=None):

		# setup logger
		if logger:
			self.logger = logger
		else:
			self.logger = self.__dummyLogger()

		self.cfg_file = cf
		self.cfg_data = {}
		self.cfg_verify = {}
		self.cfg_defaults = {}
		self.list_delim = ','

		# read config from file
		self.Read()


	class __dummyLogger():

		def debug(self, *args): pass
		info = debug
		warn = debug
		warning = debug
		error = debug
		crit = debug
		critical = debug


	def __transform(self, value, v_class):

		if v_class not in ('str', 'bool', 'int', 'float'):
			raise ConfigError(msg0106.format(v_class))

		if value == '':
			return None

		if v_class == 'str':
			return value

		elif v_class == 'bool':
			if value.lower() in ('true', 'on', 'yes', '1'):
				return True

			elif value.lower() in ('false', 'off', 'no', 0):
				return False

			else:
				raise ConfigError(msg0113.format(v_class))
		else:
			try:
				return (v_class == 'int') and int(value) or float(value)

			except ValueError:
				raise ConfigError(msg0114.format(v_class))


	def __getList(self, sect, opt, c_verify):

		# first get option as a string
		opt_str = self.config.get(sect, opt)

		# try to transform sting into a list
		self.logger.debug(msg0116.format(opt_str))
		lst = []
		for lst_el in opt_str.split(self.list_delim):

			lst_el = lst_el.strip()

			if c_verify[0] == 're':
				lst_el = self.__transform(self.__matchStr(lst_el, c_verify[1:]), 'str')

			elif c_verify[0] == 'reobj':
				lst_el = self.__makeReObj(lst_el)

			else:
				lst_el = self.__transform(lst_el, c_verify[0])

			if c_verify[0] in ("str", "int", "float"):
				self.__strInList(lst_el, c_verify)

			self.logger.debug(msg0115.format(lst_el))
			lst.append(lst_el)

		return lst


	def __getDict(self, sect, opt, c_verify):

		# first get option as a string
		opt_str = self.config.get(sect, opt)

		# try to transform sting into a dict
		self.logger.debug(msg0110.format(opt_str))
		dct={}
		for dct_el_str in opt_str.split(self.list_delim):

			# unpack dict element string
			try:
				k, v = [ e.strip() for e in dct_el_str.split(':') ]

			except ValueError:
				raise ConfigError(msg0117.format(dct_el_str))

			if c_verify[1] == 're':
				v = self.__transform(self.__matchStr(v, c_verify[2:]), 'str')

			elif c_verify[1] == 'reobj':
				v = self.__makeReObj(v)

			else:
				v = self.__transform(v, c_verify[1]),

			if c_verify[1] in ("str", "int", "float"):
				self.__strInList(v, c_verify[2:])

			dct_el = { self.__transform(k, c_verify[0]) : v }

			self.logger.debug(msg0111.format(dct_el))
			dct.update(dct_el)

		return dct


	def __matchStr(self, val, c_verify):

		# try to match sting against regex
		self.logger.debug(msg0146.format(val))
		try:
			if not c_verify[0].match(val):
				raise ConfigError(msg0150.format(val))

		except (TypeError, IndexError, AttributeError):
			# internal configuration error
			raise ConfigError(msg0147.format(val))

		return val


	def __strInList(self, val, c_verify):

		try:
			if val not in c_verify[1]:
				raise ConfigError(msg0148.format(val, c_verify[1]))

			self.logger.debug(msg0149.format(val, c_verify[1]))

		except IndexError:
			pass


	def __makeReObj(self, val):

		try:
			return re.compile(val)

		except Exception as e:
			raise ConfigError(e)


	def __chkDef(self, sect, opt):

		# raise error for unknown sections
		if sect not in self.cfg_defaults.keys():
			raise ConfigError(msg0118.format(sect))

		# raise error for unknown option
		if opt not in self.cfg_defaults[sect].keys():
			raise ConfigError(msg0103.format(sect, opt))

		return True


	def __chkVer(self, opt):

		# raise error for unknown option
		if opt not in self.cfg_verify.keys():
			raise ConfigError(msg0105.format(opt))

		return True


	def __getOpt(self, sect, opt):

		c_verify = self.cfg_verify[opt]

		# check section, option against defaults
		self.__chkDef(sect, opt)

		# list of possible value types
		val_lst = {
			"str": lambda : self.config.get(sect, opt),
			"bool" : lambda : self.config.getboolean(sect, opt),
			"int" : lambda : self.config.getint(sect, opt),
			"float" : lambda : self.config.getfloat(sect, opt),
			"list" : lambda : self.__getList(sect, opt, c_verify[1:]),
			"dict" : lambda : self.__getDict(sect, opt, c_verify[1:]),
			"re" : lambda : self.__matchStr(self.config.get(sect, opt), c_verify[1:]),
			"reobj" : lambda : self.__makeReObj(self.config.get(sect, opt)),
			}

		# convert value string for an option, section into a verified value
		# for bypass: val = self.config.get(sect, opt)
		try:
			# get value type from verify data
			val_type = c_verify[0]
			val = val_lst[val_type]()

		except KeyError:
			# option not in verify data
			raise ConfigError(msg0105.format(opt))

		except ValueError:
			# internal configuration error
			raise ConfigError(msg0119.format(sect, opt))

		except ConfigParser.NoOptionError:
			# option not in config
			self.logger.error(msg0112.format(sect, opt))
			val = None

		except ConfigParser.NoSectionError:
			# section not in config
			self.logger.debug(msg0143.format(sect))
			val = None

		except (ConfigParser.InterpolationError,
			ConfigParser.InterpolationDepthError) as e:

			raise ConfigError(msg0144.format(sect, opt, e))

		if val_type in ("str", "int", "float"):

			self.__strInList(val, c_verify)

		self.logger.debug(msg0109.format(opt, val, sect))
		return val


	def __getSect(self, sect):

		# get the sections options
		try:
			options = self.config.options(sect)

		except ConfigParser.NoSectionError:
			# section not in config
			self.logger.debug(msg0142.format(sect))
			options = []

		self.logger.debug(msg0124.format(sect, options))

		# get the options values
		opt_dct = {}
		for opt in options:
			opt_dct[opt] = self.__getOpt(sect, opt)

		return opt_dct


	def __getConf(self):

		# get config from config object
		# get config sections
		sections = self.config.sections()
		self.logger.debug(msg0120.format(sections))

		# cycle through the config files sections
		for sect in sections:
			self.cfg_data[sect] = self.__getSect(sect)

		return self.cfg_data


	def __getDefOpt(self, sect, opt):

		# add section if not in config
		if sect not in self.cfg_data.keys():
			self.logger.debug(msg0102.format(sect))
			self.cfg_data[sect] = {}

		# set option from defaults if not in config
		if opt not in self.cfg_data[sect].keys() or self.cfg_data[sect][opt] == None:
			self.cfg_data[sect][opt] = self.cfg_defaults[sect][opt]
			self.logger.debug(msg0104.format(opt, self.cfg_data[sect][opt]))
		else:
			self.logger.debug(msg0127.format(opt))


	def __getDefSect(self, sect):

		# cycle through the defaults sections options
		self.logger.debug(msg0128.format(sect))
		for opt in self.cfg_defaults[sect].keys():
			self.__getDefOpt(sect, opt)


	def Read(self):

		self.config = ConfigParser.SafeConfigParser()

		# read config from file
		try:
			if not self.config.read(self.cfg_file):
				self.logger.debug(msg0100.format(self.cfg_file))
			else:
				self.logger.debug(msg0101.format(self.cfg_file))

		except ConfigParser.MissingSectionHeaderError as e:
			# config file error
			raise ConfigError(e)


	# TODO:
	"""
	def Write(self):
		with open(self.config_file, 'wb') as cfg_file:
			self.config.write(cfg_file)


	def Export(self):
		# dict: str(d1).replace('{','').replace('}','').replace('\'','')
		# str(d1)[1:-1].replace('\'','')
		# list: str(l1).replace('(','').replace(')','').replace('\'','')
		# str(l1)[1:-1].replace('\'','')

		# add section if not in config
		if not self.config.has_section(sect):
			self.logger.debug(msg0102.format(sect))
			self.config.add_section(sect)

		# set option from defaults if not in config
		if not self.config.has_option(sect, opt):
			logger.debug(msg0004.format(opt, sect))
			self.config.set(sect, opt, str(self.cfg_defaults[sect][opt]))
	"""


	def getOption(self, sect, opt, defaults='__not_defined__', verify=None):

		# get value for an option x of section y
		# handle defaults and verify
		if defaults != '__not_defined__':
			self.updateDefaults(sect, {opt : defaults})
		if verify:
			self.updateVerify({opt : verify})

		# add section if not in config
		if sect not in self.cfg_data.keys():
			self.logger.debug(msg0102.format(sect))
			self.cfg_data[sect] = {}

		# save option value from config temporarily
		val = self.__getOpt(sect, opt)

		# store value in config data
		self.cfg_data[sect][opt] = val

		# fill empty option with default data
		self.__getDefOpt(sect, opt)

		self.logger.debug(msg0130.format(self.cfg_data[sect][opt]))
		return self.cfg_data[sect][opt]


	def getSection(self, sect, defaults=None, verify=None):

		# handle defaults and verify
		if defaults:
			self.updateDefaults(sect, defaults)
		if verify:
			self.updateVerify(verify)

		# get section from config
		self.cfg_data[sect] = self.__getSect(sect)

		# get section from defaults
		self.__getDefSect(sect)

		self.logger.debug(msg0130.format(self.cfg_data[sect]))
		return self.cfg_data[sect]


	def getConfig(self, defaults=None, verify=None):

		# handle defaults and verify
		if defaults:
			for sect in defaults.keys():
				self.updateDefaults(sect, defaults[sect])
		if verify:
			self.updateVerify(verify)

		# fetch config from config object
		self.__getConf()

		# cylce though defaults sections
		self.logger.debug(msg0107)
		for sect in self.cfg_defaults.keys():

			# fill up the section with defaults for missing options
			self.__getDefSect(sect)

		self.logger.debug(msg0130.format(self.cfg_data))
		return self.cfg_data


	def getVerify(self):

		self.logger.debug(msg0129.format(self.cfg_verify))
		return self.cfg_verify


	def setVerify(self, verify):

		self.cfg_verify = verify
		self.logger.debug(msg0108.format(self.cfg_verify))


	def flushVerify(self):

		self.cfg_verify = {}
		self.logger.debug(msg0122)


	def updateVerify(self, verify):

		try:
			self.cfg_verify.update(verify)
			self.logger.debug(msg0123.format(verify))

		except ValueError:
			self.logger.debug(msg0133.format(verify))


	def delVerify(self, opt_lst):

		for opt in opt_lst:
			try:
				del self.cfg_verify[opt]
				self.logger.debug(msg0131.format(opt))

			except KeyError:
				self.logger.warn(msg0132.format(opt))


	def getDefaults(self):

		self.logger.debug(msg0126.format(self.cfg_defaults))
		return self.cfg_defaults


	def setDefaults(self, defaults):

		self.cfg_defaults = defaults
		self.logger.debug(msg0134.format(defaults))


	def flushDefaults(self):

		self.cfg_defaults = {}
		self.logger.debug(msg0135)


	def updateDefaults(self, sect, opt_dct):

		# add section if not in defaults
		if sect not in self.cfg_defaults.keys():
			self.logger.debug(msg0121.format(sect))
			self.cfg_defaults[sect] = {}

		try:
			self.cfg_defaults[sect].update(opt_dct)

			self.logger.debug(msg0136.format(sect, opt_dct))

		except ValueError:
			self.logger.debug(msg0137.format(sect, opt_dct))


	def delDefaults(self, sect, opt_lst=None):

		if opt_lst:

			for opt in opt_lst:
				try:
					del self.cfg_defaults[sect][opt]
					self.logger.debug(msg0138.format(sect, opt))

				except KeyError:
					self.logger.warn(msg0139.format(sect, opt))

		try:
			if len(self.cfg_defaults[sect]) == 0 or not opt_lst:
				del self.cfg_defaults[sect]
				self.logger.debug(msg0140.format(sect))

		except KeyError:
			self.logger.debug(msg0141.format(sect))


	def getData(self):

		self.logger.debug(msg0130.format(self.cfg_data))
		return self.cfg_data


	def setData(self, cfg_data):

		# check all cfg_data against defaults and verify
		for sect in cfg_data.keys():
			for opt in cfg_data[sect].keys():
				self.__chkDef(sect, opt)
				self.__chkVer(opt)
				# TODO: check if the returned type of an option matches verify data
		
		self.logger.debug(msg0125.format(self.cfg_data))
		self.cfg_data = cfg_data


	def getRawSections(self):

		raw_sect = self.config.sections()
		self.logger.debug(msg0120.format(raw_sect))
		return raw_sect


	def getListDelim(self):

		return self.list_delim


	def setListDelim(self, delim):

		self.logger.debug(msg0145.format(delim))
		self.list_delim = delim

