#!/usr/bin/env python
# coding:utf-8

"""
sDSM (small Disk Space Monitor)

This file is part of sDSM.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = "Christian Schwamborn <cs@mail.architektur.tu-darmstadt.de>"
__copyright__ = "Copyright (c) 2016 Christian Schwamborn"
__license__ = "GPL v.3"
__version__ = "$Revision: 0.1 $"
__date__ = "$Date: 2016-01-26 10:16:00 $"


try:
	import sys
	import os
	import re
	import time
	from collections import namedtuple
	import signal
	import logging
	import logging.handlers
	import traceback
	import socket

	import mods.daemon as daemon
	from mods.sighandler import SignalHandler
	import mods.config as config
	import mods.smtpclient as smtpclient

except ImportError as err:

	sys.stderr.write('Module import error: {0}\n'.format(err))
	sys.exit(1)



time_re_obj = re.compile(r'^\s*(?P<num>\d+)\s*(?P<suff>[mhdw]?)\s*$')
size_re_obj = re.compile(r'^\s*(?P<num>\d+(?:.\d+)?)\s*(?P<suff>[kKmMgGtT%]?)\s*$')


def parseTime(t):

	"""calculates time with suffixes and returns seconds
	multiplicative suffixes: m, h, d, w
	"""

	suff_d = {
		'' : 1,
		'm' : 60,
		'h' : 60*60,
		'd' : 60*60*24,
		'w' : 60*60*24*7,
	}

	try:
		num, suff = time_re_obj.match(t).groups()
		t = int(num) * suff_d[suff]

	except (AttributeError, KeyError, ValueError) as e:
		raise TimeValueError(e)

	return t


def parseSize(size):

	"""calculates size with suffixes and returns bytes
	multiplicative suffixes: k, K, m, M, g, G, t, T
	small letters are powers of 1000
	big letters are powers of 1024
	"""

	suff_d = {
		'' : 1,
		'%' : -1,
		'k' : 1000,
		'K' : 1024,
		'm' : 1000*1000,
		'M' : 1024*1024,
		'g' : 1000*1000*1000,
		'G' : 1024*1024*1024,
		't' : 1000*1000*1000*1000,
		'T' : 1024*1024*1024*1024,
	}

	try:
		num, suff = size_re_obj.match(size).groups()
		size = float(num) * suff_d[suff]

	except (AttributeError, KeyError, ValueError) as e:
		raise SizeValueError(e)

	return size


def getFistEP(pl_lst):

	for pl in pl_lst:
		p = os.path.join(*pl)
		if os.path.exists(p):
			return p
	return None


def getVolumes():

	"""Return all mounted volumes as a dict of 'device : mountpt'
	"""

	try:
		with open("/proc/filesystems", "r") as pfs:

			filesystems = []
			for line in pfs:
				if not line.startswith("nodev"):
					filesystems.append(line.strip())

		with open('/proc/mounts', "r") as pmnt:

			dev_lst = {}
			for line in pmnt:

				if line.startswith('none'):
					continue

				fields = line.split()
				device = os.path.normpath(fields[0].decode('string_escape'))
				mountpt = os.path.normpath(fields[1].decode('string_escape'))
				# mountpt = bytes(fields[1], 'utf-8').decode('unicode_escape')) # python 3
				fstype = fields[2]

				if fstype not in filesystems:
					continue

				# we assume that the real mount is ahead
				# of enventual bind mounts
				if device not in dev_lst.keys():
					dev_lst[device] = mountpt

	except EnvironmentError as e:
		raise getVolumesError(e)

	return dev_lst



class TimeValueError(Exception):
	"""Indicates time value conversion errors"""
	pass


class SizeValueError(Exception):
	"""Indicates size value conversion errors"""
	pass


class getVolumesError(Exception):
	"""Indicates an error getting mounted volumes"""
	pass



class sDSMExit(Exception):

	def __init__(self, val, msg):

		self.val = val
		self.msg = msg
		sys.stderr.write('\n{0}\n'.format(self.msg))
		sys.exit(self.val)


	def __str__(self):

		return self.msg



class DiskVolumePool(object):

	def __init__(self):

		self.chk_volumes = ()


	def addVolumes(self, volumes):

		for vol in volumes.items():
			self.addVolume(vol)


	def addVolume(self, vol):

		self.chk_volumes += (DiskVolume(*vol), )


	def getVolumes(self):

		return self.chk_volumes


	def checkFree(self):

		return [ vol.checkFree() for vol in self.chk_volumes ]


	def __repr__(self):

		return "<Class:{0}>".format(self.__class__.__name__)



class DiskVolume(DiskVolumePool):

	usage = namedtuple('usage', 'path total used free percent level')

	def __init__(self, path, parm):

		self.path = path
		self.free = parseSize(parm.free)
		self.level = parm.level


	def diskUsage(self):

		"""Return disk usage associated with path"""

		st = os.statvfs(self.path.decode())
		total = (st.f_blocks * st.f_frsize)
		used = (st.f_blocks - st.f_bfree) * st.f_frsize
		free = (st.f_bavail * st.f_frsize)

		try:
			percent = ret = (float(used) / total) * 100

		except ZeroDivisionError:
			percent = 0

		# the percentage is -5% than what shown by df due to
		# reserved blocks that we are currently not considering
		# or what ever is reserved in this filesystem

		return self.usage(self.path, total, used, free, round(percent, 1), self.level)


	def checkFree(self):

		d_usage = self.diskUsage()
		size_ok = True

		if self.free > 0:
			if d_usage.free < self.free:
				size_ok = False
		else:
			if d_usage.percent + self.free > 0:
				size_ok = False

		return (size_ok, d_usage)


	def __repr__(self):

		return "<Class:{0} path:\'{1}\' free:\'{2}\>".format(self.__class__.__name__, self.path, self.free)



class sDSMDaemon(daemon.DaemonBase):

	vol_ntuple = namedtuple('volume', 'free, level')


	def __init__(self, pidfile, workpath='/'):

		super(sDSMDaemon, self).__init__(pidfile, workpath)
		self.notify = None


	def closeDaemon(self, sign):

		self.logger.info('closing Diskspace Monitor, caugth signal: {0}'.format(sign))
		logging.shutdown()
		sys.exit(0)


	def createSignalHandler(self):

		# Create SignalHandler object
		self.sighandler = SignalHandler()

		# Register a SIGTERM (kill -15) and callback
		# object when signal is trapped.
		self.sighandler.register( signal.SIGTERM, self.closeDaemon)

		# Register a SIGINT (kill -2) and callback
		# object when signal is trapped.
		# This is equivalent to CTRL-C
		self.sighandler.register( signal.SIGINT, self.closeDaemon)

		# Register a SIGHUP (kill -1) and callback
		# object when signal is trapped.
		self.sighandler.register( signal.SIGHUP, self.loadConfig)


	def getLogger(self):

		"""
		self.logger = logging.getLogger('sdsm')
		#self.logger.setLevel(logging.INFO)
		self.logger.setLevel(logging.DEBUG)

		if self.daemonized:
			handler = logging.handlers.SysLogHandler(
				address='/dev/log',
				facility=logging.handlers.SysLogHandler.LOG_SYSLOG)
			formatter = logging.Formatter('%(name)s[%(process)d]: ' +
				'%(levelname)s - %(message)s')
		else:
			handler = logging.StreamHandler()
			formatter = logging.Formatter('%(asctime)s - %(levelname)s ' + 
				'[%(module)s/%(funcName)s/%(lineno)d] %(message)s')
			self.logger.setLevel(logging.DEBUG)

		handler.setFormatter(formatter)
		self.logger.addHandler(handler)
		"""

		if not self.daemonized:
			c_handler = logging.StreamHandler()
			c_formatter = logging.Formatter('%(asctime)s - %(levelname)s ' + 
				'[%(module)s/%(funcName)s/%(lineno)d] %(message)s')
			c_handler.setFormatter(c_formatter)
			c_handler.setLevel(logging.DEBUG)
			logger.addHandler(c_handler)
			#logger.setLevel(logging.DEBUG)

		return logger


	def loadConfig(self, sign=None):

		if sign == 1:
			self.logger.info('Reload Config ...')
		else:
			self.logger.info('Loading Config ...')

		self.report_timestamps = {}
		all_volumes = getVolumes()

		cfg_ver = {
			'scannall' : ('bool', ),
			'warntosyslog' : ('bool', ),
			'path' : ('str', ),
			'free' : ('re', size_re_obj),
			'level' : ('str', ('user', 'admin', '', ), ),
			'exclude' : ('list', 'reobj'),
			'testinterval' : ('re', time_re_obj),
			'reportinterval' : ('re', time_re_obj),
			'debug' : ('bool', ),
			'host' : ('str', ),
			'port' : ('str', ),
			'security' : ('str', ('ssl', 'tls', 'none', '', ), ),
			'user' : ('str', ),
			'pwd' : ('str', ),
			'sender' : ('str', ),
			'adminmail' : ('list', 'str'),
			'usermail' : ('list', 'str'),
		}

		g_cfg_def = {
			'scannall' : True,
			'warntosyslog' : True,
			'free' : '90%',
			'level' : 'admin',
			'exclude' : [],
			'testinterval' : '10m',
			'reportinterval' : '1d',
			'debug' : False,
			'adminmail' : [],
			'usermail' : [],
		}

		smtp_cfg_def = {
			'host' : '',
			'port' : '',
			'security' : '',
			'user' : '',
			'pwd' : '',
			'sender' : '',
		}

		a_cfg_def = {
			'path' : '',
			'free' : '',
			'level' : '',
		}

		# determin the config file
		config_file = getFistEP( (
			(sys.path[0], 'etc', 'sdsm_dev.conf'),
			(sys.path[0], 'etc', 'sdsm.conf'),
			(os.path.sep, 'etc', 'sdsm.conf'),
			(os.path.sep, 'dev', 'null'),
		) )

		if not config_file:
			logging.shutdown()
			raise sDSMExit(1, "unable to find a configuration file (or /dev/null)")

		# read and validate config
		try:
			cfg_obj = config.Config(config_file, self.logger)
			cfg_obj.setListDelim(';')
			cfg_obj.updateVerify(cfg_ver)
			cfg_obj.updateDefaults('Global', g_cfg_def)
			cfg_obj.updateDefaults('SMTP', smtp_cfg_def)

			# set defaults for all non ('Global', 'SMTP') sections
			for sect in cfg_obj.getRawSections():

				if sect not in ('Global', 'SMTP'):
					cfg_obj.updateDefaults(sect, a_cfg_def)
					cfg_sect = cfg_obj.getSection(sect)

		except config.ConfigError as e:
			loggering.shutdown()
			raise sDSMExit(1, e)

		# get final config
		cfg = cfg_obj.getConfig()
		self.cfg_global = cfg.pop('Global')
		self.cfg_smtp = cfg.pop('SMTP')

		# build list of volumes to monitor
		volumes = {}
		for vol in cfg.values():

			path = os.path.normpath(vol['path'])

			if os.path.ismount(path) and \
				path in all_volumes.values():

				# path is already a mount point
				mnt = path
			else:
				try:
					# get mount point for device
					mnt = all_volumes[os.path.realpath(path)]

				except KeyError:
					self.logger.error('unable to lookup mounted volume: \'{0}\''.format(path))
					continue

			volumes[mnt] = self.vol_ntuple(
				vol['free'] or self.cfg_global['free'],
				vol['level'] or self.cfg_global['level'])

		for vol, mnt in all_volumes.items():

			# add all mounted volumed if scannall = True and not already in list
			if self.cfg_global['scannall'] and \
				mnt not in volumes.keys():

				volumes[mnt] = self.vol_ntuple(
					self.cfg_global['free'], self.cfg_global['level'])

			# apply possible excludes
			for excl in self.cfg_global['exclude']:
				if excl.match(vol) or excl.match(mnt):
					self.logger.debug('remove \'{0} : {1}\' from list'.format(vol, mnt))
					del(volumes[mnt])
					continue

		# create volume pool to check
		self.dvp = DiskVolumePool()
		self.dvp.addVolumes(volumes)


	def run(self):

		# setup signal handler
		self.createSignalHandler()

		# setup logging
		self.logger = self.getLogger()
		self.logger.info('Diskspace Monitor started ...')

		# initialize config
		self.loadConfig()

		# apply debug logging
		if self.cfg_global['debug']:
			#logger.setLevel(logging.DEBUG)
			s_handler.setLevel(logging.DEBUG)

		# init smtp mail object
		smtp = smtpclient.SmtpMail(logger=self.logger, **self.cfg_smtp)

		while True:

			for vol in self.dvp.checkFree():

				self.logger.debug(vol)
				size_ok = vol[0]
				usage = vol[1]

				if size_ok:

					if usage.path in self.report_timestamps.keys():

						# remove time stamp of volumes below threshold
						del(self.report_timestamps[usage.path])
						self.logger.debug('volume \'{0}\' is now below threshold again: \'{1}\''.format(usage.path, usage.percent))

					continue

				# if size is not ok we may report

				# time stamps are in UTC
				ct_s = time.time()

				# TODO:	collect all mail, send at the end and
				#	reset timestamps for all full volumes
				if usage.path not in self.report_timestamps.keys():

					self.logger.debug('volume \'{0}\' is over threshold: \'{1}\''.format(usage.path, usage.percent))

				elif ct_s < self.report_timestamps[usage.path] + \
					parseTime(self.cfg_global['reportinterval']):

					self.logger.debug('skipping report for \'{0}\''.format(usage.path))
					continue

				self.report_timestamps[usage.path] = ct_s
				msg = 'volume \'{0}\' runs out of space, currently at {1}%'.format(usage.path, usage.percent)
				subj = 'Disk space warning from \'{0}\''.format(socket.getfqdn())

				if usage.level == 'admin':

					log_level = self.logger.warn
					mail_addr = self.cfg_global['adminmail']
				else:
					log_level = self.logger.info
					mail_addr = self.cfg_global['usermail']

				if self.cfg_global['warntosyslog'] or \
					not self.daemonized:

					log_level(msg)

				if self.cfg_smtp['host'] and mail_addr:
					smtp.send(subj, msg, recipient=mail_addr)

			time.sleep(parseTime(self.cfg_global['testinterval']))



def main():

	usage = 'Missing parameter, usage:\n' + \
			' % sdsmd start|restart|stop|foreground\n'

	if len(sys.argv) < 2:
		raise sDSMExit(2, usage)

	# setup an early logger to syslog for debuging
	global logger
	global s_handler
	logger = logging.getLogger('sdsm')
	logger.setLevel(logging.DEBUG)
	s_handler = logging.handlers.SysLogHandler(
		address='/dev/log',
		facility=logging.handlers.SysLogHandler.LOG_SYSLOG)
	s_formatter = logging.Formatter('%(name)s[%(process)d]: ' +
		'%(levelname)s - %(message)s')
	s_handler.setFormatter(s_formatter)
	s_handler.setLevel(logging.INFO)
	logger.addHandler(s_handler)

	# place pid file, this might be very debian specific
	if os.getuid() == 0:
		pid_file_path = getFistEP( (
			(os.path.sep, 'run'),
			(os.path.sep, 'var', 'run'),
			(os.path.sep, 'tmp'),
		) )
	else:
		pid_file_path = getFistEP( (
			(os.path.sep, 'tmp'),
		))

	if not pid_file_path:
		logger.error('unable to find a pid file path')
		logging.shutdown()
		raise sDSMExit(1, 'unable to find a pid file path')

	pid_file = os.path.join(pid_file_path, 'sdsmd.pid')


	try:
		dc = daemon.DaemonCtl(sDSMDaemon, pid_file)

		if sys.argv[1] == 'start':
			dc.start()

		elif sys.argv[1] == 'stop':
			dc.stop()

		elif sys.argv[1] == 'restart':
			dc.restart()

		elif sys.argv[1] == 'foreground':
			dc.foreground()

		else:
			sys.stderr.write('\n{0}\n'.format(usage))
			sys.exit(2)

	except Exception as e:

		logger.error(traceback.format_exc())
		# this doesn't report to stderr, as all stream handlers has been
		# disconnected during daemonization
		logging.shutdown()
		raise sDSMExit(1, e)



if __name__ == '__main__':

	main()

